/*
 * University of Illinois Open Source License
 * Copyright 2011-2013 Luthey-Schulten Group,
 * All rights reserved.
 *
 * Developed by: Luthey-Schulten Group
 *               University of Illinois at Urbana-Champaign
 *               http://www.scs.uiuc.edu/~schulten
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the Software), to deal with
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to
 * do so, subject to the following conditions:
 *
 * - Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimers.
 *
 * - Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimers in the documentation
 * and/or other materials provided with the distribution.
 *
 * - Neither the names of the Luthey-Schulten Group, University of Illinois at
 * Urbana-Champaign, nor the names of its contributors may be used to endorse or
 * promote products derived from this Software without specific prior written
 * permission.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * THE CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR
 * OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS WITH THE SOFTWARE.
 *
 * Author(s): Andrew Magis, Elijah Roberts
 */

#ifndef LM_BUILDER_CUBOID_H_
#define LM_BUILDER_CUBOID_H_

#include "lm/builder/Shape.h"
#include "lm/rdme/Lattice.h"

namespace lm {
namespace builder {

/// @class Cuboid
/// @brief A cube-like Shape
class Cuboid : public Shape
{
public:
    /// @brief Create a Cuboid that is axis aligned
    /// @param p1 Point of the lower corner
    /// @param p2 Point of the upper corner
    /// @param type The type of the sites within the cuboid
    Cuboid(point p1, point p2, site_t type);
	/// @brief Create a Cuboid based on two orientations, and a width, height, depth
	/// @brief center The definition of the center of the cuboid
	/// @param w Width (along x axis when unrotated)
	/// @param h Height (along y axis when unrotated)
	/// @param d Depth (along z axis when unrotated)
	/// @param at A vector representing where the cuboid is pointing (default along x axis)
	/// @param at A vector representing where the cuboid up is (default along y axis)
	Cuboid(point center, si_dist_t w, si_dist_t h, si_dist_t d, site_t type, vector at = vector(1.0,0.0,0.0), vector up = vector(0.0,1.0,0.0));

    /// @brief Destroy the Cuboid
    virtual ~Cuboid();
    
    /// @brief Check if two shapes intersect
    /// @param query The other shape to check
    /// @return true/false
    virtual bool intersects(Shape * query);
    /// @brief Determine if the shape contains the specified point
    /// @param query Point to test
    /// @return true/false
    virtual bool contains(point query);
    /// @brief Determine if the shape contains the specified shape
    /// @param query Shape to test
    /// @return true/false
    virtual bool contains(Shape * query);
    /// @brief Get the lower point
    virtual point getP1() {return p1;}
    /// @brief Get the upper point
    virtual point getP2() {return p2;}
    /// @brief Get the volume bounded by the cuboid shape
    virtual double getVolume();

protected:
    point p1, p2;
	point cen;

	si_dist_t width;
	si_dist_t height;
	si_dist_t depth;
};

}
}

#endif

