/*
 * University of Illinois Open Source License
 * Copyright 2011-2015 Luthey-Schulten Group,
 * All rights reserved.
 *
 * Developed by: Luthey-Schulten Group
 *               University of Illinois at Urbana-Champaign
 *               http://www.scs.uiuc.edu/~schulten
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the Software), to deal with
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to
 * do so, subject to the following conditions:
 *
 * - Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimers.
 *
 * - Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimers in the documentation
 * and/or other materials provided with the distribution.
 *
 * - Neither the names of the Luthey-Schulten Group, University of Illinois at
 * Urbana-Champaign, nor the names of its contributors may be used to endorse or
 * promote products derived from this Software without specific prior written
 * permission.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * THE CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR
 * OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS WITH THE SOFTWARE.
 *
 * Author(s): Joseph R. Peterson
 */

#include "lm/builder/Shape.h"
#include "lm/builder/Sphere.h"
#include "lm/builder/Cylinder.h"
#include "lm/rdme/Lattice.h"

namespace lm {
    namespace builder {
        
        Cylinder::Cylinder(point p1, point p2, si_dist_t radius, site_t type)
        :Shape(CYLINDER,Cylinder::calcBoundingBox(p1,p2,radius), type, vector(p1,p2).unitize(), vector::findPerpendicularVector(vector(p1,p2)).unitize()),p1(p1),p2(p2),radius(radius),length(p1.distance(p2))
        {
        }
        
        bounding_box Cylinder::calcBoundingBox(point p1, point p2, si_dist_t radius)
        {
            bounding_box bbox;
            bbox.min.x = (p1.x<p2.x)?(p1.x):(p2.x);
            bbox.max.x = (p1.x>p2.x)?(p1.x):(p2.x);
            bbox.min.y = (p1.y<p2.y)?(p1.y):(p2.y);
            bbox.max.y = (p1.y>p2.y)?(p1.y):(p2.y);
            bbox.min.z = (p1.z<p2.z)?(p1.z):(p2.z);
            bbox.max.z = (p1.z>p2.z)?(p1.z):(p2.z);
            return bbox;
        }
        
        Cylinder::~Cylinder()
        {
        }
        
        bool Cylinder::intersects(Shape * query)
        {
            // FIXME: NEEDS TO BE IMPLEMENTED //
            if (!boundingBoxesIntersect(query)) return false;
            
            if (query->getShapeType() == SPHERE)
            {
                return true;
            }
            return false;
        }
        
        bool Cylinder::contains(point query)
        {
            // Transform the point as if the capsule was oriented in the +z direction with p1 at 0,0,r and p2 at 0,0,l-r.
            
            // Assume for now that the capsule is oriented with its long axis along the z axis.
            
            // Translate the sphere in the x-y plane.
            query.x -= p1.x;
            query.y -= p1.y;
            query.z -= p1.z;
            
            // Check if point is inside
            si_dist_t radius2 = radius*radius;
            if (query.z < length && query.z > 0.0) // Cylinder region
            {
                if (query.x*query.x + query.y*query.y > radius2)
                    return false;
                else
                    return true;
            }
			return false;
        }
        
        bool Cylinder::contains(Shape * query)
        {
            if (!boundingBoxesIntersect(query)) return false;
            
            
            if (query->getShapeType() == SPHERE)
            {
                Sphere * querySphere = (Sphere *)query;
                
                // Transform the sphere as if the capsule was oriented in the +z direction with p1 at 0,0,r and p2 at 0,0,l-r.
                point tc = querySphere->getCenter();
                si_dist_t tr=querySphere->getRadius();
                
                // Assume for now that the capsule is oriented with its long axis along the z axis.
                
                // Translate the sphere in the x-y plane.
                tc.x -= p1.x;
                tc.y -= p1.y;
                tc.z -= p1.z;
                
                // See if the sphere is closer to the lower cap, the cylinder, or the upper cap.
                if (tc.z > length-radius)
                {
                    if (tc.x*tc.x+tc.y*tc.y+(tc.z-length)*(tc.z-length) > (radius-tr)*(radius-tr)) return false;
                    return true;
                }
            }
            return false;
        }
        
        double Cylinder::getVolume()
        {
            return PI*radius*radius*length;
        }
        
        
    }
}
